/*
 * Created by SharpDevelop.
 * User: grunwald
 * Date: 25.07.2006
 * Time: 09:14
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */

using System;
using System.Collections;
using PPJ.Runtime.Scripting;

namespace ScaeParser
{
	public class SALParser : IScriptParser
	{
		private ScaeLexer lexer;

		public ASTNode Parse(string code)
		{
			this.lexer = new ScaeLexer(code);

			if (string.Compare(code, 0, "set ", 0, 4, true) == 0)
				return ParseAssignment();
			else
				return ParseExpression();
		}

		internal Expression ParseExpression()
		{
			Expression n = Expr();
			Expect(TokenKind.Eof, Errors.ExpectEndOfExpression);
			return n;
		}

		internal Assignment ParseAssignment()
		{
			Expression left = SimpleExpr();
			Expect(TokenKind.Equals, Errors.ExpectAssign);
			Expression right = Expr();
			Expect(TokenKind.Eof, Errors.ExpectEndOfExpression);
			return new Assignment(left, right, lexer.Location);
		}

		TokenKind la
		{
			get
			{
				return lexer.NextToken.Kind;
			}
		}

		void Next()
		{
			lexer.ReadNextToken();
		}

		void Expect(TokenKind k, string errorText)
		{
			if (lexer.ReadNextToken().Kind != k)
			{
				throw lexer.Error(errorText);
			}
		}

		Expression Expr()
		{
			return LogicalOrExpr();
		}

		Expression LogicalOrExpr()
		{
			Expression n = LogicalAndExpr();
			while (la == TokenKind.Or)
			{
				Next();
				n = new Binary(BinaryOperator.LogicalOr, n, LogicalAndExpr(), lexer.Location);
			}
			return n;
		}

		Expression LogicalAndExpr()
		{
			Expression n = LogicalNotExpr();
			while (la == TokenKind.And)
			{
				Next();
				n = new Binary(BinaryOperator.LogicalAnd, n, LogicalNotExpr(), lexer.Location);
			}
			return n;
		}

		Expression LogicalNotExpr()
		{
			if (la == TokenKind.Not)
			{
				Next();
				return new Unary(UnaryOperator.LogicalNot, RelationExpr(), lexer.Location);
			}
			else
			{
				return RelationExpr();
			}
		}

		Expression RelationExpr()
		{
			Expression n = AdditiveExpr();
			BinaryOperator op = GetRelationOperator(la);
			while (op != BinaryOperator.As) // As = None
			{
				Next();
				n = new Binary(op, n, AdditiveOrNotExpr(), lexer.Location);
				op = GetRelationOperator(la);
			}
			return n;
		}

		BinaryOperator GetRelationOperator(TokenKind k)
		{
			switch (k)
			{
				case TokenKind.Equals:
					return BinaryOperator.Equality;
				case TokenKind.Inequals:
					return BinaryOperator.Inequality;
				case TokenKind.LessThan:
					return BinaryOperator.LessThan;
				case TokenKind.GreaterThan:
					return BinaryOperator.GreaterThan;
				case TokenKind.LessThanOrEqual:
					return BinaryOperator.LessThanOrEqual;
				case TokenKind.GreaterThanOrEqual:
					return BinaryOperator.GreaterThanOrEqual;
				default:
					return BinaryOperator.As;
			}
		}

		Expression AdditiveOrNotExpr()
		{
			if (la == TokenKind.Not)
			{
				Next();
				return new Unary(UnaryOperator.LogicalNot, AdditiveExpr(), lexer.Location);
			}
			else
			{
				return AdditiveExpr();
			}
		}

		Expression AdditiveExpr()
		{
			Expression n = MultiplicativeExpr();
			TokenKind k = la;
			while (k == TokenKind.Plus || k == TokenKind.Minus || k == TokenKind.StringConcat)
			{
				Next();
				if (k == TokenKind.Minus)
				{
					n = new Binary(BinaryOperator.Subtraction, n, MultiplicativeExpr(), lexer.Location);
				}
				else
				{
					n = new Binary(BinaryOperator.Addition, n, MultiplicativeExpr(), lexer.Location);
				}
				k = la;
			}
			return n;
		}

		Expression MultiplicativeExpr()
		{
			Expression n = BitwiseExpr();
			TokenKind k = la;
			while (k == TokenKind.Multiplication || k == TokenKind.Division)
			{
				Next();
				if (k == TokenKind.Multiplication)
				{
					n = new Binary(BinaryOperator.Multiply, n, BitwiseExpr(), lexer.Location);
				}
				else
				{
					n = new Binary(BinaryOperator.Division, n, BitwiseExpr(), lexer.Location);
				}
				k = la;
			}
			return n;
		}

		Expression BitwiseExpr()
		{
			Expression n = MinusExpr();
			TokenKind k = la;
			while (k == TokenKind.BitAnd || k == TokenKind.BitOr)
			{
				Next();
				if (k == TokenKind.BitAnd)
				{
					n = new Binary(BinaryOperator.BitwiseAnd, n, MinusExpr(), lexer.Location);
				}
				else
				{
					n = new Binary(BinaryOperator.BitwiseOr, n, MinusExpr(), lexer.Location);
				}
				k = la;
			}
			return n;
		}

		Expression MinusExpr()
		{
			if (la == TokenKind.Minus)
			{
				Next();
				return new Unary(UnaryOperator.Negation, SimpleExpr(), lexer.Location);
			}
			else
			{
				return SimpleExpr();
			}
		}

		// SimpleExpr ::= ( "(" Expr ")" | literal | identifier )
		//                { "." identifier | "(" ArgList ")" | "[" ArgList "]"  }
		Expression SimpleExpr()
		{
			Expression n;
			if (la == TokenKind.Literal)
			{
				n = lexer.ReadNextToken().LiteralValue;
			}
			else if (la == TokenKind.Identifier)
			{
				n = new SimpleName(lexer.ReadNextToken().IdentifierName, lexer.Location);
			}
			else if (la == TokenKind.OpenParenthesis)
			{
				Next();
				n = Expr();
				Expect(TokenKind.CloseParenthesis, Errors.ExpectCloseParenthesis);
			}
			else
			{
				throw lexer.Error(Errors.ExpectLiteralOrVariable, la);
			}
			TokenKind k = la;
			while (k == TokenKind.OpenParenthesis || k == TokenKind.Dot || k == TokenKind.OpenBracket)
			{
				Next();
				if (k == TokenKind.Dot)
				{
					string ident = lexer.NextToken.IdentifierName;
					Expect(TokenKind.Identifier, Errors.ExpectIdentifierAfterDot);
					n = new MemberAccess(n, ident, lexer.Location);
				}
				else
				{
					TokenKind endToken = (k == TokenKind.OpenParenthesis) ? TokenKind.CloseParenthesis : TokenKind.CloseBracket;
					ArrayList arguments = new ArrayList();
					if (la != endToken)
					{
						if (k == TokenKind.OpenBracket)
						{
							arguments.Add(Expr());
							while (la == TokenKind.Comma)
							{
								Next();
								arguments.Add(Expr());
							}
						}
						else
						{
							arguments.Add(new Argument(Expr(), ArgumentType.In, lexer.Location));
							while (la == TokenKind.Comma)
							{
								Next();
								arguments.Add(new Argument(Expr(), ArgumentType.In, lexer.Location));
							}
						}
					}

					if (k == TokenKind.OpenParenthesis)
					{
						Expect(endToken, Errors.ExpectCommaOrCloseParenthesis);
						n = new Invocation(n, arguments, lexer.Location);
					}
					else
					{
						Expect(endToken, Errors.ExpectCommaOrCloseBracket);
						n = new ElementAccess(n, arguments, lexer.Location);
					}
				}
				k = la;
			}
			return n;
		}
	}
}
